/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#ifndef __math_Vector_h__
#define __math_Vector_h__

namespace fe
{

/**************************************************************************//**
	@brief Dense vector - size fixed by template

	@ingroup math
*//***************************************************************************/
template<int N, class T>
class Vector
{
	public:
					Vector(void);
					template<typename U>
					Vector(const Vector<N,U> &other);
					template<typename U>
					Vector(const Vector<N+1,U> &other);
					template<typename U>
					Vector(const Vector<N-1,U> &other);
					Vector(const T *array);
					Vector(T x,T y,T z=(T)0,T w=(T)0);
//					Vector(const T &all);

const	T&			operator[](unsigned int index) const;
		T&			operator[](unsigned int index);

		T			*raw(void);
const	T			*raw(void) const;

					// like raw, but only valid if used on the call
const	T			*temp(void) const;

		Vector<N,T>	&operator=(const Vector<N,T> &other);
		template<typename U>
		Vector<N,T>	&operator=(const Vector<N,U> &other);
		Vector<N,T>	&operator=(const Vector<N+1,T> &other);
		template<typename U>
		Vector<N,T>	&operator=(const Vector<N+1,U> &other);
		template<typename U>
		Vector<N,T>	&operator=(const Vector<N-1,U> &other);
		Vector<N,T>	&operator=(const T *array);

		template<typename U>
		Vector<N,T>	&operator=(const Vector<N-2,U> &other);

		bool		operator==(const Vector<N,T> &other);
		bool		operator!=(const Vector<N,T> &other);

	public:
		T	m_data[N];
};


template<int N, class T>
inline Vector<N,T>::Vector(void)
{
	for(unsigned int i = 0; i < N; i++)
	{
		m_data[i] = T(0);
	}
}

template<int N, class T>
template<typename U>
inline Vector<N,T>::Vector(const Vector<N,U> &other)
{
	for(unsigned int i = 0; i < N; i++)
	{
		m_data[i] = (T)other[i];
	}
}

template<int N, class T>
template<typename U>
inline Vector<N,T>::Vector(const Vector<N+1,U> &other)
{
	for(unsigned int i = 0; i < N; i++)
	{
		m_data[i] = other[i];
	}
}

template<int N, class T>
template<typename U>
inline Vector<N,T>::Vector(const Vector<N-1,U> &other)
{
	for(unsigned int i = 0; i < N-1; i++)
	{
		m_data[i] = other[i];
	}
	if(N>1)
	{
		m_data[N-1]=T(0);
	}
}

template<int N, class T>
inline Vector<N,T>::Vector(const T *array)
{
	operator=(array);
}

template<int N,class T>
inline Vector<N,T>::Vector(T x, T y, T z, T w)
{
	m_data[0] = x;
	if(N >=2) m_data[1] = y;
	if(N >=3) m_data[2] = z;
	if(N >=4) m_data[3] = w;
}

#if FALSE
template<int N,class T>
inline Vector<N,T>::Vector(const T &all)
{
	for(unsigned int i = 0; i < N; i++)
	{
		m_data[i] = all;
	}
}
#endif

template<int N, class T>
inline const T& Vector<N,T>::operator[](unsigned int index) const
{
#if FE_BOUNDSCHECK
	if(index >= N) { feX(e_invalidRange); }
#endif
	return m_data[index];
}

template<int N, class T>
inline T &Vector<N,T>::operator[](unsigned int index)
{
#if FE_BOUNDSCHECK
	if(index >= N) { feX(e_invalidRange); }
#endif
	return m_data[index];
}

template<int N, class T>
inline T *Vector<N,T>::raw(void)
{
	return m_data;
}

template<int N, class T>
inline const T *Vector<N,T>::raw(void) const
{
	return m_data;
}

template<int N, class T>
inline const T *Vector<N,T>::temp(void) const
{
	return m_data;
}

template<int N, class T>
inline bool Vector<N,T>::operator==(const Vector<N,T> &other)
{
	if(this != &other)
	{
		for(unsigned int i = 0; i < N; i++)
		{
			if(m_data[i] != other[i]) { return false; }
		}
	}
	return true;
}

template<int N, class T>
inline bool Vector<N,T>::operator!=(const Vector<N,T> &other)
{
	return !operator==(other);
}

template<int N, class T>
inline Vector<N,T> &Vector<N,T>::operator=(const Vector<N,T> &other)
{
	if(this != &other)
	{
		for(unsigned int i = 0; i < N; i++)
		{
			m_data[i] = other[i];
		}
	}
	return *this;
}

template<int N, class T>
template<typename U>
inline Vector<N,T> &Vector<N,T>::operator=(const Vector<N,U> &other)
{
	if(this != &other)
	{
		for(unsigned int i = 0; i < N; i++)
		{
			m_data[i] = (T)other[i];
		}
	}
	return *this;
}

template<int N, class T>
inline Vector<N,T> &Vector<N,T>::operator=(const Vector<N+1,T> &other)
{
	for(unsigned int i = 0; i < N; i++)
	{
		m_data[i] = other[i];
	}
	return *this;
}

template<int N, class T>
template<typename U>
inline Vector<N,T> &Vector<N,T>::operator=(const Vector<N+1,U> &other)
{
	for(unsigned int i = 0; i < N; i++)
	{
		m_data[i] = (T)other[i];
	}
	return *this;
}

template<int N, class T>
template<typename U>
inline Vector<N,T> &Vector<N,T>::operator=(const Vector<N-1,U> &other)
{
	for(unsigned int i = 0; i < N-1; i++)
	{
		m_data[i] = (T)other[i];
	}
	if(N>1)
	{
		m_data[N-1]=T(0);
	}
	return *this;
}

template<int N, class T>
template<typename U>
inline Vector<N,T>& Vector<N,T>::operator=(const Vector<N-2,U> &other)
{
	for(unsigned int i = 0; i < N-2; i++)
	{
		m_data[i] = (T)other[i];
	}
	if(N>1)
	{
		m_data[N-1]=T(0);
		if(N>2)
		{
			m_data[N-2]=T(0);
		}
	}
	return *this;
}

template<int N, class T>
inline Vector<N,T> &Vector<N,T>::operator=(const T *array)
{
	for(unsigned int i = 0; i < N; i++)
	{
		m_data[i] = array[i];
	}
	return *this;
}

/* non member operations */

/** Set all the elements to the given value
	@relates Vector
	*/
template<int N, class T, class U>
inline Vector<N,T>& setAll(Vector<N,T> &lhs,const U value)
{
	for(unsigned int i = 0; i < N; i++)
	{
		lhs[i] = value;
	}
	return lhs;
}

/** Set the value at the index
	@relates Vector
	*/
template<int N, class T,class U>
inline Vector<N,T> &setAt(Vector<N,T> &lhs, U32 index,const U value)
{
	lhs[index]=value;
	return lhs;
}

/** Return the number of elements
	@relates Vector
	*/
template<int N, class T>
inline U32 size(const Vector<N,T> &lhs)
{
	return lhs.size();
}

/** In place add operator
	@relates Vector
	*/
template<int N, class T>
inline Vector<N,T> &operator+=(Vector<N,T> &lhs, const Vector<N,T> &rhs)
{
	for(unsigned int i = 0; i < N; i++)
	{
		lhs[i] += rhs[i];
	}
	return lhs;
}

/** In place add operator
	@relates Vector
	*/
template<int M, int N, class T>
inline Vector<N,T> &operator+=(Vector<N,T> &lhs, const Vector<M,T> &rhs)
{
	const unsigned int limit=(M<N)? M: N;
	for(unsigned int i = 0; i < limit; i++)
	{
		lhs[i] += rhs[i];
	}
	return lhs;
}

/** In place subtract operator
	@relates Vector
	*/
template<int N, class T>
inline Vector<N,T> &operator-=(Vector<N,T> &lhs, const Vector<N,T> &rhs)
{
	for(unsigned int i = 0; i < N; i++)
	{
		lhs[i] -= rhs[i];
	}
	return lhs;
}

/** In place subtract operator
	@relates Vector
	*/
template<int M, int N, class T>
inline Vector<N,T> &operator-=(Vector<N,T> &lhs, const Vector<M,T> &rhs)
{
	const int limit=(M<N)? M: N;
	for(unsigned int i = 0; i < limit; i++)
	{
		lhs[i] -= rhs[i];
	}
	return lhs;
}

/** Negate operation
	@relates Vector
	*/
template<int N, class T>
inline Vector<N,T> operator-(const Vector<N,T> &rhs)
{
	Vector<N,T> v;
	for(unsigned int i = 0; i < N; i++)
	{
		v[i] = -rhs[i];
	}
	return v;
}

/** In place piecewise multiply operator
	@relates Vector
	*/
template<int N, class T>
inline Vector<N,T> &operator*=(Vector<N,T> &lhs, const Vector<N,T> &rhs)
{
	for(unsigned int i = 0; i < N; i++)
	{
		lhs[i] = lhs[i] * rhs[i];
	}
	return lhs;
}

/** In place piecewise multiply operator
	@relates Vector
	*/
template<int M, int N, class T>
inline Vector<N,T> &operator*=(Vector<N,T> &lhs, const Vector<M,T> &rhs)
{
	const int limit=(M<N)? M: N;
	for(unsigned int i = 0; i < limit; i++)
	{
		lhs[i] *= rhs[i];
	}
	return lhs;
}

/** In place piecewise scale operator
	@relates Vector
	*/
template<int N, class T,class U>
inline Vector<N,T> &operator*=(Vector<N,T> &lhs, U scale)
{
	for(unsigned int i = 0; i < N; i++)
	{
		lhs[i] *= scale;
	}
	return lhs;
}

/** Dot (inner) product
	@relates Vector
	*/
template<int N, class T>
inline T dot(const Vector<N,T> &lhs, const Vector<N,T> &rhs)
{
	T t = (T)0;
	for(unsigned int i = 0; i < N; i++)
	{
		t += lhs[i] * rhs[i];
	}
	return t;
}

/** Frobenius norm operation
	@relates Vector
	*/
template<int N, class T>
inline T magnitude(const Vector<N,T> &rhs)
{
	T mag = 0.0;
	for(unsigned int i = 0; i < N; i++)
	{
		mag += (T)rhs[i] * (T)rhs[i];
	}
	return sqrt(mag);
}

/** Square of the length
	@relates Vector
	*/
template<int N, class T>
inline T magnitudeSquared(const Vector<N,T> &rhs)
{
	T mag = 0.0;
	for(unsigned int i = 0; i < N; i++)
	{
		mag += (T)rhs[i] * (T)rhs[i];
	}
	return mag;
}

/** Return vector scaled to unit length
	@relates Vector
	*/
template<int N, class T>
inline Vector<N,T> unit(const Vector<N,T> &vec)
{
	T mag=magnitude(vec);
	if(mag==T(0))
	{
		feX("unit(Vector<N,T>)","attempt to normalize zero magnitude vector");
	}
	T inv=T(1)/mag;
	Vector<N,T> v;
	for(U32 i=0;i<N;i++)
	{
		v[i]=inv*vec[i];
	}
	return v;
}

/** Return vector scaled to unit length with zero check
	@relates Vector
	*/
template<int N, class T>
inline Vector<N,T> unitSafe(const Vector<N,T> &vec)
{
	T mag=magnitude(vec);
	if(mag>T(0))
	{
		T inv=T(1)/mag;
		Vector<N,T> v;
		for(U32 i=0;i<N;i++)
		{
			v[i]=inv*vec[i];
		}
		return v;
	}
	return vec;
}

/** In place normalize operator
	@relates Vector
	*/
template<int N, class T>
inline Vector<N,T> &normalize(Vector<N,T> &vec)
{
	T mag=magnitude(vec);
	if(mag==T(0))
	{
		feX("Vector<N,T>::normal",
			"attempt to normalize zero magnitude vector");
	}
	T inv=T(1)/mag;
	for(U32 i=0;i<N;i++)
	{
		vec[i]*=inv;
	}
	return vec;
}

/** In place normalize operator with zero length check
	@relates Vector
	*/
template<int N, class T>
inline Vector<N,T> &normalizeSafe(Vector<N,T> &vec)
{
	T mag=magnitude(vec);
	if(mag>T(0))
	{
		T inv=T(1)/mag;
		for(U32 i=0;i<N;i++)
		{
			vec[i]*=inv;
		}
	}
	return vec;
}

/** Print to a string
	@relates Vector
	*/
template<int N, class T>
inline String print(const Vector<N,T> &vec)
{
	String s = "";
	for(unsigned int i = 0; i < N; i++)
	{
		s.cat("%.6G", vec[i]);
		if(i < N-1)
		{
			s.cat(" ");
		}
	}
	return s;
}

/** add operation
	@relates Vector
	*/
template<int N, class T>
inline Vector<N,T> operator+(const Vector<N,T> &lhs, const Vector<N,T> &rhs)
{
	Vector<N,T> v;
	for(unsigned int i = 0; i < N; i++)
	{
		v[i] = lhs[i] + rhs[i];
	}
	return v;
}

/** add operation
	@relates Vector
	*/
template<int N, int M, class T>
inline Vector<N,T> operator+(const Vector<N,T> &lhs, const Vector<M,T> &rhs)
{
	Vector<N,T> v;
	for(unsigned int i = 0; i < N; i++)
	{
		if(i >= M)
		{
			v[i] = lhs[i];
		}
		else
		{
			v[i] = lhs[i] + rhs[i];
		}
	}
	return v;
}


/** subtract operation
	@relates Vector
	*/
template<int N, class T>
inline Vector<N,T> operator-(const Vector<N,T> &lhs, const Vector<N,T> &rhs)
{
	Vector<N,T> v;
	for(unsigned int i = 0; i < N; i++)
	{
		v[i] = lhs[i] - rhs[i];
	}
	return v;
}

/** subtract operation
	@relates Vector
	*/
template<int N, int M, class T>
inline Vector<N,T> operator-(const Vector<N,T> &lhs, const Vector<M,T> &rhs)
{
	Vector<N,T> v;
	for(unsigned int i = 0; i < N; i++)
	{
		if(i >= M)
		{
			v[i] = lhs[i];
		}
		else
		{
			v[i] = lhs[i] - rhs[i];
		}
	}
	return v;
}

/** equality test
	@relates Vector
	*/
template<int N, class T>
inline bool operator==(const Vector<N,T> &lhs, const Vector<N,T> &rhs)
{
	for(unsigned int i = 0; i < N; i++)
	{
		if(rhs[i] != lhs[i])
		{
			return false;
		}
	}
	return true;
}

/** Equivalence test within the given tolerance @em margin
	@relates Vector
	*/
template<int N, class T,class U>
inline bool equivalent(const Vector<N,T> &lhs, const Vector<N,T> &rhs, U margin)
{
	for(unsigned int i = 0; i < N; i++)
	{
		if(FE_INVALID_SCALAR(lhs[i]) ||
				FE_INVALID_SCALAR(rhs[i]) ||
				fabs(rhs[i] - lhs[i]) > margin)
		{
			return false;
		}
	}
	return true;
}

/** Piecewise multiply operation
	@relates Vector
	*/
template<int N, class T>
inline Vector<N,T> operator*(const Vector<N,T> &lhs, const Vector<N,T> &rhs)
{
	Vector<N,T> v;
	for(unsigned int i = 0; i < N; i++)
	{
		v[i] = lhs[i] * rhs[i];
	}
	return v;
}

/** Scale operation
	@relates Vector
	*/
template<int N, class T,class U>
inline Vector<N,T> operator*(const U lhs, const Vector<N,T> &rhs)
{
	Vector<N,T> v;
	for(unsigned int i = 0; i < N; i++)
	{
		v[i] = rhs[i] * lhs;
	}
	return v;
}

/** @brief Scale operation

	@relates Vector
	*/
template<int N, class T,class U>
inline Vector<N,T> operator*(const Vector<N,T> &lhs, const U &rhs)
{
	Vector<N,T> v;
	for(unsigned int i = 0; i < N; i++)
	{
		v[i] = lhs[i] * rhs;
	}
	return v;
}

/** @brief Inverse Scale operation

	@relates Vector
	*/
template<int N, class T,class U>
inline Vector<N,T> operator/(const Vector<N,T> &lhs, const U &rhs)
{
	if(rhs==0.0f)
		feX(e_impossible,"operator/(Vector<N,T>,T)","divide by zero");
	return lhs*(1.0f/rhs);
}

/** @brief Add with scaling

	lhs = lhs + rhs * scalar

	@relates Vector
	*/
template<int N, class T,class U>
inline Vector<N,T> &addScaled(Vector<N,T> &lhs,U scalar,
		const Vector<N,T> &rhs)
{
	for(unsigned int i = 0; i < N; i++)
	{
		lhs[i]+=scalar*rhs[i];
	}
	return lhs;
}

/** @brief Scale then add

	lhs = lhs * scalar + rhs

	@relates Vector
	*/
template<int N, class T,class U>
inline Vector<N,T> &scaleAndAdd(Vector<N,T> &lhs,U scalar,
		const Vector<N,T> &rhs)
{
	for(unsigned int i = 0; i < N; i++)
	{
		lhs[i]=lhs[i]*scalar+rhs[i];
	}
	return lhs;
}

} /* namespace */

#endif /* __math_Vector_h__ */

