/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#include "surface.pmh"

#define	FE_ST_TREE_DRAW_DEBUG	TRUE
#define	FE_ST_TREE_DRAW_NORMALS	FALSE

namespace fe
{

SurfaceTriangles::SurfaceTriangles(void)
{
	m_pImpactPool=&m_triangleImpactPool;
}

SurfaceTriangles::~SurfaceTriangles(void)
{
}

SpatialTransform SurfaceTriangles::sample(I32 a_face,
	SpatialBary a_barycenter) const
{
	const I32 face3=a_face*3;
	if(face3+3>m_vertices)
	{
		feLog("SurfaceTriangles::sample excess face %d/%d\n",
				a_face,m_vertices/3);
		return SpatialTransform::identity();
	}
	const SpatialVector& rVertex0=m_pVertexArray[face3];
	const SpatialVector& rVertex1=m_pVertexArray[face3+1];
	const SpatialVector& rVertex2=m_pVertexArray[face3+2];
	const SpatialVector& rNormal0=m_pNormalArray[face3];
	const SpatialVector& rNormal1=m_pNormalArray[face3+1];
	const SpatialVector& rNormal2=m_pNormalArray[face3+2];

	const SpatialVector point=location(a_barycenter,rVertex0,rVertex1,rVertex2);
	SpatialVector norm=location(a_barycenter,rNormal0,rNormal1,rNormal2);
	Real mag=magnitude(norm);
	SpatialTransform result;
	if(mag>0.0)
	{
		makeFrameNormalY(result,point,rVertex1-rVertex0,norm/mag);
	}
	else
	{
		setIdentity(result);
		setTranslation(result,point);
	}

#if FALSE
	feLog("face %d/%d bary %s\n",a_face,m_vertices/3,c_print(a_barycenter));
	feLog("rVertex0 %s\n",c_print(rVertex0));
	feLog("rVertex1 %s\n",c_print(rVertex1));
	feLog("rVertex2 %s\n",c_print(rVertex2));
	feLog("rNormal0 %s\n",c_print(rNormal0));
	feLog("rNormal1 %s\n",c_print(rNormal1));
	feLog("rNormal2 %s\n",c_print(rNormal2));
	feLog("point %s\n",c_print(point));
	feLog("norm %s\n",c_print(norm));
	feLog("result %s\n",c_print(result));
#endif

	return result;
}

void SurfaceTriangles::drawInternal(BWORD a_transformed,
	const SpatialTransform* a_pTransform,sp<DrawI> a_spDrawI,
	const fe::Color* a_pColor) const
{
//	return;

#if FE_ST_TREE_DRAW_DEBUG
	sp<DrawMode> spWireframe(new DrawMode);
	spWireframe->copy(a_spDrawI->drawMode());
	spWireframe->setDrawStyle(DrawMode::e_outline);
	a_spDrawI->pushDrawMode(spWireframe);
#endif

#if TRUE
	if(m_vertices)
	{
		const Color white(1.0f,1.0f,1.0f,1.0f);

		const Vector2f* pTexture=NULL;
		const BWORD multicolor=FALSE;
		if(a_transformed)
		{
			a_spDrawI->drawTransformedTriangles(*a_pTransform,NULL,
					m_pVertexArray,m_pNormalArray,pTexture,m_vertices,
					DrawI::e_discrete,multicolor,a_pColor? a_pColor: &white);
		}
		else
		{
			a_spDrawI->drawTriangles(m_pVertexArray,m_pNormalArray,pTexture,
					m_vertices,DrawI::e_discrete,multicolor,
					a_pColor? a_pColor: &white);
		}
	}
	else
	{
		feLog("SurfaceTriangles::draw no vertices\n");
	}
#endif

#if FE_ST_TREE_DRAW_NORMALS
	if(m_vertices)
	{
		const Color red(1.0f,0.0f,0.0f,1.0f);
		const Real normalScale=1.0;
		SpatialVector pLines[2];
		for(U32 m=0;m<m_vertices;m++)
		{
			pLines[0]=m_pVertexArray[m];
			pLines[1]=pLines[0]+m_pNormalArray[m]*normalScale;

			const SpatialVector* pNormals=NULL;
			const BWORD multicolor=FALSE;
			if(a_transformed)
			{
				a_spDrawI->drawTransformedLines(*a_pTransform,NULL,pLines,
						pNormals,2,DrawI::e_discrete,multicolor,&red);
			}
			else
			{
				a_spDrawI->drawLines(pLines,pNormals,2,DrawI::e_discrete,
						multicolor,&red);
			}
		}
	}
#endif

#if FE_ST_TREE_DRAW_DEBUG
	a_spDrawI->popDrawMode();
#endif

#if FE_ST_TREE_DRAW_DEBUG
	if(m_spSpatialTreeI.isValid())
	{
		m_spSpatialTreeI->draw(a_spDrawI,NULL);
	}
#endif
}

void SurfaceTriangles::Impact::copy(sp<SpatialTreeI::Hit>& a_rspHit)
{
	SurfaceSearchable::Impact::copy(a_rspHit);
	setFace(a_rspHit->face());
	setBarycenter(a_rspHit->barycenter());
	setVertex0(a_rspHit->vertex()[0]);
	setVertex1(a_rspHit->vertex()[1]);
	setVertex2(a_rspHit->vertex()[2]);
	setNormal0(a_rspHit->normal()[0]);
	setNormal1(a_rspHit->normal()[1]);
	setNormal2(a_rspHit->normal()[2]);
}

std::vector< sp<SurfaceI::ImpactI> > SurfaceTriangles::nearestPoints(
		const SpatialVector& a_origin,U32 a_hitLimit) const
{
	std::vector< sp<SurfaceI::ImpactI> > impactArray=
			SurfaceSearchable::nearestPoints(a_origin,a_hitLimit);

	//* is this neccessary?
	const U32 hitCount=impactArray.size();
	for(U32 m=0;m<hitCount;m++)
	{
		sp<Impact> spImpact=m_triangleImpactPool.get();
		spImpact->setSurface(dynamic_cast<const SurfaceTriangles*>(this));
	}

	return impactArray;
}

sp<SurfaceI::ImpactI> SurfaceTriangles::nearestPoint(
		const SpatialVector& a_origin) const
{
	sp<SurfaceTriangles::Impact> spImpact=
			SurfaceSearchable::nearestPoint(a_origin);

	if(spImpact.isValid())
	{
		const SpatialVector normal=spImpact->normal();
		spImpact->setNormal0(normal);
		spImpact->setNormal1(normal);
		spImpact->setNormal2(normal);
	}

	return spImpact;
}

std::vector< sp<SurfaceI::ImpactI> > SurfaceTriangles::rayImpacts(
		const SpatialVector& a_origin,const SpatialVector& a_direction,
		Real a_maxDistance,U32 a_hitLimit) const
{
	std::vector< sp<SurfaceI::ImpactI> > impactArray=
			SurfaceSearchable::rayImpacts(a_origin,a_direction,
			a_maxDistance,a_hitLimit);

	//* is this neccessary?
	const U32 hitCount=impactArray.size();
	for(U32 m=0;m<hitCount;m++)
	{
		sp<Impact> spImpact=m_triangleImpactPool.get();
		spImpact->setSurface(dynamic_cast<const SurfaceTriangles*>(this));
	}

	return impactArray;
}

sp<SurfaceI::ImpactI> SurfaceTriangles::rayImpact(
		const SpatialVector& a_origin,const SpatialVector& a_direction,
		Real a_maxDistance) const
{
	sp<SurfaceTriangles::Impact> spImpact=
			SurfaceSearchable::rayImpact(a_origin,a_direction,a_maxDistance);

	//* is this neccessary?
	if(spImpact.isValid())
	{
		spImpact->setSurface(dynamic_cast<const SurfaceTriangles*>(this));
	}

	return spImpact;
}

void SurfaceTriangles::resolveImpact(sp<ImpactI> a_spImpactI) const
{
	sp<Impact> spImpact(a_spImpactI);

	// TEMP if no nearest
	if(!spImpact.isValid() || spImpact->distance()<Real(0))
	{
		return;
	}

	SpatialVector intersection;
	SpatialVector normal;
	RayTriangleIntersect<Real>::resolveContact(
			spImpact->vertex0(),
			spImpact->vertex1(),
			spImpact->vertex2(),
			spImpact->normal0(),
			spImpact->normal1(),
			spImpact->normal2(),
			spImpact->origin(),
			spImpact->direction(),
			spImpact->barycenter(),
			spImpact->distance(),
			intersection,normal);
	spImpact->setIntersectionLocal(intersection);
	spImpact->setNormalLocal(normal);
}

} // namespace
